/* A prototype compiler for CLAIR.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

:- dynamic(currentSymbolNum/2).

clair_compile(X) :-
    current_output(OldStream),
    open('clairCompiled.mc', write, Stream),
    set_output(Stream),
    write('void compiledCLAIR() {'), nl,
    compE([], 0, X),
    write('}'),
    close(Stream),
    set_output(OldStream).

cbop(eq, 'EQ').
cbop(ne, 'NE').
cbop(lt, 'LT').
cbop(le, 'LE').
cbop(ge, 'GE').
cbop(gt, 'GT').

cbop(plus,  'PLUS').
cbop(minus, 'MINUS').
cbop(or,    'OR').

cbop(prod, 'PROD').
cbop(div,  'DIV').
cbop(mod,  'MOD').
cbop(and,  'AND').

emit(Opc) :-
    write(' '), write(Opc), nl.

emit(Opc, X) :-
    write(' '), write(Opc), write('('), write(X), write(')'), nl.

emit(Opc, X, Y) :-
    write(' '), write(Opc), write('('), write(X), write(', '),
    write(Y), write(')'), nl.

emitLab(L) :-
    write(L), write(': ;'), nl.

compE(Env, N, i(X)) :-
    getEnv(Env, X, M, Offset),
    NLinks is N-M,
    emit('VAR', NLinks, Offset).

compE(_, _, tt) :-
    emit('PUSH', 1).

compE(_, _, ff) :-
    emit('PUSH', 0).

compE(_, _, X) :-
    integer(X), emit('PUSH', X).

compE(Env, N, let(D, E)) :-
    dDV(D, V),
    M is N+1,
    assign(V, M, Env1),
    length(V, L),
    emit('FRAME', L, -1),
    compD(Env, N, Env1, D, Env2),
    compE(Env2, M, E),
    emit('DROP', 1).

compE(Env, N, X) :-
    X =.. [O, E1, E2], cbop(O, M),
    compE(Env, N, E1),
    compE(Env, N, E2), emit(M).

compE(Env, N, if(E, E1, E2)) :-
     newLabel(L1), newLabel(L2),
     compE(Env, N, E),
     emit('GOFALSE', L1),
     compE(Env, N, E1),
     emit('GO', L2),
     emitLab(L1),
     compE(Env, N, E2),
     emitLab(L2).

newLabel(L) :-
    newSymbol(l, L).

newSymbol(Root, S) :-
    retract(currentSymbolNum(Root, Num1)), !,
    Num is Num1+1,
    asserta(currentSymbolNum(Root, Num)),
    name(Num, LNum),
    name(Root, LRoot),
    append(LRoot, LNum, L),
    name(S, L).

newSymbol(Root, S) :-
    asserta(currentSymbolNum(Root, 0)),
    name(Root, LRoot),
    append(LRoot, [48], L),
    name(S, L).


getEnv([e(X, N, O)|_], X, N, O) :- !.
getEnv([_|T], X, N, O) :-
    getEnv(T, X, N, O).

assign(V, N, Env1) :-
    assign1(V, N, Env1, 0).

assign1([X | V], N, [e(X, N, Offset) | Env], Offset) :-
    Offset1 is Offset+1,
    assign1(V, N, Env, Offset1).
assign1([], _, [], _).

compD(Env, N, Env1, def(X, _, E), [e(X, N, Offset)]) :-
    compE(Env, N, E),
    getEnv(Env1, X, M, Offset),
    emit('STORE', M, Offset).

compD(Env, N, Env1, seq(D1, D2), Env2) :-
    compD(Env, N, Env1, D1, Env3),
    compose(Env, Env3, Env4),
    compD(Env4, N, Env1, D2, Env5),
    compose(Env3, Env5, Env2).
